package ch.hevs.kart;

/**
 * Enumeration in order to identify the different control registers of the Kart.
 * <br><br>
 * Control registers of the kart can be read or written. Some of them contain hardware configuration
 * that has only be initialized once like {@link DrivePwmPeriod}, {@link SteeringStepPeriod},
 * {@link SteeringEndSwitchAddress}, {@link HardwareSettings} and  {@link UpdateInterval} - some
 * of them control the kart directly like the steering position control register
 * {@link SteeringPosition}, the drive control register {@link DriveSpeed} or the digital
 * outputs/LED register {@link LEDs}. Others might be the front-end of a FIFO that writes to the
 * kart memory like the {@link SequenceFifoRegister}.
 * <br><br>
 * This class implements the protocol/register map specification from
 * <a href="http://wiki.hevs.ch/fsi/index.php5/Kart/serial_link">here</a>
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
public enum KartControlRegister {
	/**
	 * The Drive PWM period as UINT16 (0..65535).
     * <br><br>
     * Divider value from the FPGA main clock frequency to the PWM step frequency. Default value is
     * 0x40 = 64. This means that the FPGA clock is divided by 64.
	 */
	DrivePwmPeriod(0),

	/**
	 * Drive motor speed as INT5 (-15..15).
     * <br><br>
     * Controls the actual moving speed of the kart, whereas 0 means stop. Negative numbers from -1
     * to -15 will move the kart backwards, -15 is the highest speed backwards and -1 the slowest.
     * Positive numbers from 1 to 15 will move the kart forwards, 1 is the slowest speed and 15 the
     * highest.
	 */
	DriveSpeed(1),

	/**
	 * The Stepper (Steering) motor step period as UINT16.
     * <br><br>
     * Divider value from the clock frequency to the stepper motor step period.
     * Default value is 0x100 = 256. This means that the FPGA clock is divided by 256.
	 */
	SteeringStepPeriod(2),

	/**
	 * Steering position set point as UINT16 register.
     * <br><br>
     * Controls the steering stepper motor target position, whereas zero is at the stepper end
     * switch position. Values are positive only and the center and maximal position depend the
     * kart geometry.
	 */
	SteeringPosition(3),

	/**
	 * Specifies the address at which the end switch is read at bit 0 (if it isn't local).
	 */
	SteeringEndSwitchAddress(4),

	/**
	 * Hardware settings register. Defines how the motor rotations are linked to the abstract kart
     * move actions.
     * <br><br>
	 *
     * <ul>
	 *     <li>
	 *         <b>Bit 0:</b> Kart goes as DC motor.
	 *         <br>
	 *         When '0', the kart goes backwards when the motor turns forward.
	 *     </li>
	 *     <li>
	 *         <b>Bit 1:</b> Kart turns right (clockwise).
	 *         <br>
	 *         When '1', kart turns to the right as stepper motor phases go from 1 to 4.
	 *     </li>
	 *	   <li>
	 *	       <b>Bit 2:</b> End sensor is on left side.
	 *	       <br>
	 *	       If '1', the angles are measured clockwise.
	 *	   </li>
	 *	   <li>
	 *	       <b>Bit 3:</b> Stepper end emulation.
	 *	       <br>
	 *	       Emulates an end switch contact for the direction motor.
	 *	   </li>
	 *	   <li>
	 *	       <b>Bit 4:</b> Restart slaves.
	 *	       <br>
	 *	       Puts following slaves back in startup mode: stepper motor.
	 *	   </li>
     * </ul>
	 */
	HardwareSettings(5),

	/**
	 * LED control register.
     * <br><br>
     * Allows to change the state of the 4 on-board LED/digital output drivers.
	 */
	LEDs(6),

    /**
     * Allows to control the sequence virtual machine on the kart.
     * <br><br>
     * There are currently three supported commands that will be executed directly at the moment of
     * the write to the register:
     * <br><br>
     * <ul>
     *  <li>
     *      <b>0x0000:</b> Stop sequence, switch to idle mode (back to real-time control).
     *  </li>
     *  <li>
     *      <b>0x0001:</b> Reset/stop sequence virtual machine and remove the current sequence from
     *      the kart's memory.
     *  </li>
     *  <li>
     *      <b>0x0002:</b> Start/restart the actual sequence in the karts virtual machine memory.
     *      This sequence has to be downloaded using the SequenceFifoRegister before issuing this
     *      command.
     *  </li>
     * </ul>
     * <br><br>
     * You should never use the method {@link Kart#setControlRegister(KartControlRegister, int)} to
     * change the value of this register, as the Kart class automatically synchronizes this
     * registers to the kart and might issue the command multiple times. The Kart class has an API
     * to download and run sequences: {@link Kart#run(KartSequence)}.
     * <br><br>
     * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
     */
	SequenceControlRegister(7),

    /**
     * Allows to download a sequence to the kart's virtual machine memory. Each write operation to
     * this register will write a new line to the sequence memory on the kart.
     * <br><br>
     * You should never use the method {@link Kart#setControlRegister(KartControlRegister, int)} to
     * download a sequence to the kart, as the Kart class automatically synchronizes this registers
     * to the kart and will most probably send the same sequence instruction multiple times. The
     * Kart class has an API to download and run sequences: {@link Kart#run(KartSequence)}.
     * <br><br>
     * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
     */
	SequenceFifoRegister(8),

	/**
	 * Update interval (from kart to App) in milliseconds.
     * <br><br>
     * This control register sets the amount of time between each synchronisation of the status
     * registers from the kart to the Android application. Default value is 100.
	 */
	UpdateInterval(15);

	KartControlRegister(final int address) {
		this.address = address;
	}

	/**
	 * Returns the address of the register in the kart's memory.
	 *
	 * @return Address of the register.
	 */
	public int getAddress() {
		return address;
	}

	// Maximal count of registers.
	static final int COUNT = 16;

	// The address of the register.
	private final int address;
}
