package ch.hevs.kart;

import android.content.Context;
import android.os.Bundle;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import ch.hevs.utils.Limits;

/**
 * A KartSequence is a step-by-step sequence of commands that can be downloaded to the Kart.
 * <br><br>
 * You can see the hardware based sequence player on the Kart as a very simple virtual machine that
 * executes one instruction after another and supports a well defined set of opcodes along with
 * their arguments. This objects abstracts the hard work to generate these vm commands by exposing a
 * simple interface to construct the sequence and using the method {@link Kart#run(KartSequence)}
 * you can execute such a sequence on the Kart.
 * <br><br>
 * The builder pattern ({@link KartSequence.Builder}) can be used in order to construct such a
 * sequence. Sequences can be saved and loaded to/from xml files by using the methods
 * {@link KartSequence#save} and {@link KartSequence#load}.
 * <br><br>
 * <b>IMPORTANT: This feature is experimental: This means that there might be still bugs and the
 * sequence player might not be implemented on all Karts!</b>
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
public class KartSequence {
    /**
     * Creates and returns a new Builder in order to construct a Kart sequence. See the
     * {@link KartSequence.Builder} documentation for available commands a sequence can be made of.
     *
     * @return  New KartSequence Builder.
     */
    static public Builder Builder() {
        return new Builder();
    }

    /**
     * Builder that must be used in order to construct new KartSequences. Every method returns a
     * reference to the builder himself, so it is possible to create sequences in just one line of
     * code.
     *
     * <code>
     *     KartSequence seq = KartSequence.Builder().steer(0).drive(3).during(5000);
     * </code>
     *
     * The example above lets the Kart drive in a circle for about 3 seconds.
     */
    public static class Builder {
        /**
         * This instruction does actually nothing. It is present for completeness.
         *
         * @return  Reference to the builder object.
         */
        public Builder nop() {
            sequenceToConstruct.addInstruction(Instruction.OpCode.NOP, (short)0);
            return this;
        }

        /**
         * This instruction changes the speed of the kart. Valid values for the speed are -15..15,
         * whereas 0 means stop, -15 is the fastest speed backwards and 15 is the fastest speed in
         * the forward direction.
         *
         * @param speed Speed to set (-15..15).
         * @return      Reference to the builder object.
         */
        public Builder drive(int speed) {
            speed = Limits.limit(speed, Kart.MINIMAL_SPEED, Kart.MAXIMAL_SPEED);
            sequenceToConstruct.addInstruction(Instruction.OpCode.SET_SPEED, (short)speed);
            return this;
        }

        /**
         * This instruction changes the steering to the given position. Valid values for the
         * command are 0..4095, but the highest value depends the actual geometry of the Kart.
         * There is no control at all if the given parameter is compatible with the steering
         * mechanism, so it is your duty to ensure that the values are acceptable.
         *
         * @param position  Position to set steering (0..4095, but depends steering geometry).
         * @return          Reference to the builder object.
         */
        public Builder steer(int position) {
            position = Limits.limit(position, 0, 4095);
            sequenceToConstruct.addInstruction(Instruction.OpCode.SET_ANGLE, (short)position);
            return this;
        }

        /**
         * This instruction sets the four LED outputs to the values of the four least significant
         * bytes of the given value. All other bits are ignored.
         *
         * @param values    Value to set the four outputs (LED) to.
         * @return          Reference to the builder object.
         */
        public Builder outputs(int values) {
            values = values & 0xF;
            sequenceToConstruct.addInstruction(Instruction.OpCode.DRIVE_LEDS, (short)values);
            return this;
        }

        /**
         * This instruction will turn the LED at the given index (0..3) on or generally spoken, the
         * digital output will be activated.
         *
         * @param index Index of the LED/digital output to turn on (0..4).
         * @return      Reference to the builder object.
         */
        public Builder ledOn(final int index) {
            if (index >= 0 && index < 4) {
                sequenceToConstruct.addInstruction(Instruction.OpCode.SET_LED, (short)index);
                return this;
            } else {
                throw new IndexOutOfBoundsException(
                        "There are only four LEDs/outputs on the kart (0-3)");
            }
        }

        /**
         * This instruction will turn the LED at the given index (0..3) off or generally spoken, the
         * digital output will be deactivated.
         *
         * @param index Index of the LED/digital output to turn off (0..4).
         * @return      Reference to the builder object.
         */
        public Builder ledOff(final int index) {
            if (index >= 0 && index < 4) {
                sequenceToConstruct.addInstruction(Instruction.OpCode.CLEAR_LED, (short)index);
                return this;
            } else {
                throw new IndexOutOfBoundsException(
                        "There are only four LEDs/outputs on the kart (0-3)");
            }
        }

        /**
         * This instruction will leave the Kart in the exact same state until the given amount of
         * pulses are counted on hall sensor 1. If the hall sensor counter register corresponds to
         * the number of rotations the axis has made, you can make the kart traveling a certain
         * distance with the current steering position and speed. The number of instructions can be between
         * 0 and 4095.
         * <br><br>
         * Setting the speed {@link KartSequence.Builder#drive(int)} to 0 before this instruction
         * might not be a smart move...
         *
         * @param pulseCount    Number of hall sensor 1 pulses to wait until the next action
         *                      (0..4095).
         * @return              Reference to the builder object.
         */
        public Builder travel(int pulseCount) {
            pulseCount = Limits.limit(pulseCount, 0, 4095);
            sequenceToConstruct.addInstruction(Instruction.OpCode.RUN_DISTANCE, (short)pulseCount);
            return this;
        }

        /**
         * This instruction will leave the Kart in the exact same state for the given amount of
         * milliseconds. This method of waiting is certainly less exact in terms of the absolute
         * position than the {@link KartSequence.Builder#travel(int)} instruction, but might be
         * still useful for simple tasks.
         * <br><br>
         * Note that the actual resolution of the times is 10ms. A delay of 0ms up to about 41s
         * can be achieved by providing values from 0 (0ms) to 40950 (40.95s). If the passed value
         * is bigger than 40950, the delay will be limited to 41s.
         *
         * @param milliseconds  Number of milliseconds to wait before executing the next
         *                      instruction.
         * @return              Reference to the builder object.
         */
        public Builder during(int milliseconds) {
            milliseconds = Limits.limit(milliseconds, 0, 40950);
            sequenceToConstruct.addInstruction(Instruction.OpCode.RUN_FOR,
                    (short)(milliseconds / 10));
            return this;
        }

        // TODO: Document...
        public Builder registerSet(short value) {
            value = Limits.limit(value, (short)0, (short)1023);
            sequenceToConstruct.addRegisterInstruction(
                    RegisterInstruction.RegisterOpCode.REGISTER_SET, value);
            return this;
        }

        // TODO: Document...
        public Builder registerAdd(short value) {
            value = Limits.limit(value, (short)0, (short)1023);
            sequenceToConstruct.addRegisterInstruction(
                    RegisterInstruction.RegisterOpCode.REGISTER_ADD, value);
            return this;
        }

        // TODO: Document...
        public Builder registerSub(short value) {
            value = Limits.limit(value, (short)0, (short)1023);
            sequenceToConstruct.addRegisterInstruction(
                    RegisterInstruction.RegisterOpCode.REGISTER_SUB, value);
            return this;
        }

        /**
         * This will add a GOTO_IF_REG_NZ instruction to the sequence. The operand is the absolute
         * index of the instruction in the sequence to which the GOTO_IF_REG_NZ instruction should
         * point. Once the Kart sequence VM executes this instruction, it will jump to the
         * instruction that is pointed by the operand.
         *
         * @param sequenceIndex Index of the instruction to jump to.
         * @return              Reference to the builder object.
         */
        public Builder jump(final int sequenceIndex) {
            if (sequenceIndex >= 0 && sequenceIndex < sequenceToConstruct.size()) {
                sequenceToConstruct.addInstruction(Instruction.OpCode.GOTO_IF_REG_NZ, (short)sequenceIndex);
            } else {
                throw new IndexOutOfBoundsException("Jump destination is outside sequence!");
            }
            return this;
        }

        /**
         * Creates and returns the actual kart sequence. Note that the builder can be reused in
         * order to create a new sequence.
         *
         * @return  Built kart sequence.
         */
        public KartSequence build() {
            KartSequence sequence = sequenceToConstruct;
            sequenceToConstruct = new KartSequence();
            return sequence;
        }

        /**
         * Returns the count of instructions in the sequence that is actually under construction
         * by the builder.
         *
         * @return  Number of instructions in the sequence.
         */
        public int sequenceSize() {
            return sequenceToConstruct.size();
        }

        // Only KartSequence class can instantiate a builder object.
        private Builder() {}

        // The actual sequence in construction.
        private KartSequence sequenceToConstruct = new KartSequence();
    }

    /**
     * Instruction class, represents a single instruction (opcode + operand).
     */
    public static class Instruction {
        /**
         * Enumerates the opcodes actually supported by the Kart'a sequence vm.
         */
        public enum OpCode {
            /**
             * NOP. Probably not used that often...
             */
            NOP((byte)0),

            /**
             * Set the speed of the drive subsystem.
             */
            SET_SPEED((byte)1),

            /**
             * Set the steering position.
             */
            SET_ANGLE((byte)2),

            /**
             * Set digital output port to the given value (4 lsb ar considered).
             */
            DRIVE_LEDS((byte)3),

            /**
             * Turn the LEDs on according to the given mask.
             */
            SET_LED((byte)4),

            /**
             * Turn the LEDs off according to the given mask.
             */
            CLEAR_LED((byte)5),

            /**
             * Wait for the given amount of hall pulses on hall sensor 1.
             */
            RUN_DISTANCE((byte)8),

            /**
             * Wait for the given amount of milliseconds before executing the next instruction.
             */
            RUN_FOR((byte)9),

            /**
             * Modifies the register according to the given sub-opcode and and operand. The register
             * can be used in conjunction with the GOTO_IF_REG_NZ instruction which is only executed
             * if the register is non-zero.
             */
            MODIFY_REGISTER((byte)0xD),

            /**
             * Go to a given index in the sequence.
             */
            GOTO_IF_REG_NZ((byte)0xE),

            /**
             * End of the sequence (Added by the builder automatically) and starts the sequence.
             */
            END((byte)0xF);

            OpCode(final byte opCode) {
                this.opCode = opCode;
            }

            // Converts the opcode to a byte that can be interpreted by the kart VM.
            public byte toByte() {
                return opCode;
            }

            private byte opCode;
        }

        /**
         * A complete kart instruction consisting of an opcode and an operand.
         *
         * @param opCode    Kart instruction opcode (command).
         * @param operand   Kart instruction operand (argument).
         */
        public Instruction(final OpCode opCode, final short operand) {
            this.opCode = opCode;
            this.operand = operand;
        }

        /**
         * Returns true if the instruction is valid, false otherwise.
         *
         * @return  True if valid.
         */
        boolean isValid() {
            // FIXME: should check here if the operand of each instruction is valid or not
            // (for instance check if the goto operand is valid)
            return opCode != null && operand >= -15 && operand < 4096;
        }

        /**
         * Returns the opcode (command) of the instruction.
         *
         * @return  Opcode.
         */
        public OpCode getOpCode() {
            return opCode;
        }

        /**
         * Returns the operand (argument) of the instruction.
         *
         * @return  Operand.
         */
        public short getOperand() {
            return operand;
        }

        /**
         * Changes the operand (argument) of the instruction. Note that the values are limited
         * automatically to the range 0..4095.
         *
         * @param operand   New operand to set for the instruction.
         */
        public void setOperand(short operand) {
            this.operand = operand;
        }

        // Convert the instruction to a raw Kart VM instruction.
        int getKartVMInstruction() {
            int instruction = opCode.toByte() << 12;
            instruction |= operand & 0x0FFF;
            return instruction;
        }

        /**
         * Converts the instruction to an XML tag that can be embedded into XML files.
         *
         * @return  XML tag representing the instruction.
         */
        public String toXmlTag() {
            return String.format("<Instruction opcode=\"%s\" operand=\"%s\"/>",
                    opCode.toString(), operand);
        }

        @Override
        public String toString() {
            return String.format("%s:%d", opCode.toString(), operand);
        }

        /**
         * Tries to convert a string into an instruction. The instruction string should have the
         * format "{OPCODE}:{OPERAND}". Returns null if the string could not be converted into an
         * instruction.
         *
         * @param string    String to convert the instruction from.
         * @return          Instruction converted from the string or null if the string is invalid.
         */
        public static Instruction fromString(final String string) {
            if (string != null) {
                String[] tokens = string.split(":");
                if (tokens.length == 2) {
                    try {
                        OpCode opcode = OpCode.valueOf(tokens[0]);
                        short operand = Short.parseShort(tokens[1]);
                        return new Instruction(opcode, operand);
                    } catch (IllegalArgumentException e) {
                        return null;
                    }
                } else if (tokens.length == 3) {
                    try {
                        OpCode opcode = OpCode.valueOf(tokens[0]);
                        if (opcode == OpCode.MODIFY_REGISTER) {
                            RegisterInstruction.RegisterOpCode registerOpcode =
                                    RegisterInstruction.RegisterOpCode.valueOf(tokens[1]);
                            short operand = Short.parseShort(tokens[2]);
                            return new RegisterInstruction(registerOpcode, operand);
                        } else {
                            return null;
                        }
                    } catch (IllegalArgumentException e) {
                        return null;
                    }
                }
            }
            return null;
        }

        private OpCode opCode;
        private short operand;
    }

    // TODO: Document.
    public static class RegisterInstruction extends Instruction {
        public enum RegisterOpCode {
            REGISTER_SET((byte)0x0),
            REGISTER_ADD((byte)0x2),
            REGISTER_SUB((byte)0x3);

            RegisterOpCode(final byte opCode) {
                this.opCode = opCode;
            }

            public byte toByte() {
                return opCode;
            }

            private byte opCode;
        }

        public RegisterInstruction(final RegisterOpCode registerOpCode, short operand) {
            super(OpCode.MODIFY_REGISTER, (short)(operand & 0x3F));
            this.registerOpCode = registerOpCode;
        }

        public RegisterOpCode getRegisterOpcode() {
            return registerOpCode;
        }

        public void setRegisterOpcode(RegisterOpCode registerOpcode) {
            this.registerOpCode = registerOpcode;
        }

        @Override
        public void setOperand(short operand) {
            super.setOperand((short)(operand & 0x3FF));
        }

        @Override
        int getKartVMInstruction() {
            int instruction = super.getKartVMInstruction();
            instruction |= registerOpCode.toByte() << 10;
            return instruction;
        }

        @Override
        public String toXmlTag() {
            return String.format("<Instruction opcode=\"%s\" registerOpcode=\"%s\" operand=\"%s\"/>",
                    getOpCode().toString(), registerOpCode.toString(), getOperand());
        }

        @Override
        public String toString() {
            return String.format("%s:%s:%d", getOpCode().toString(), registerOpCode.toString(),
                    getOperand());
        }

        private RegisterOpCode registerOpCode;
    }

    /**
     * Returns the number of instructions in the sequence.
     *
     * @return  Number of instructions.
     */
    public int size() {
        return instructions.size();
    }

    /**
     * Returns a reference to the instruction at the given index in order to modify the instruction
     * or null if the index is outside the sequence bounds.
     *
     * @param index Index of the instruction in the sequence.
     * @return      Reference to the instruction.
     */
    public Instruction getInstruction(final int index) {
        if (index >= 0 && index < instructions.size()) {
            return instructions.get(index);
        } else {
            return null;
        }
    }

    /**
     * Adds a new instruction with the given opcode and operand to the end of the sequence.
     *
     * @param opcode    Instruction's opcode.
     * @param operand   Instruction's operand.
     */
    public void addInstruction(final Instruction.OpCode opcode, final short operand) {
        addInstruction(new Instruction(opcode, operand));
    }

    // TODO: Document.
    public void addRegisterInstruction(final RegisterInstruction.RegisterOpCode opcode,
                                       final short operand) {
        addInstruction(new RegisterInstruction(opcode, operand));
    }

    /**
     * Adds the given instruction to the end of the sequence.
     *
     * @param instruction   Instruction to add to the end of the sequence.
     */
    public void addInstruction(final Instruction instruction) {
        if (instruction != null && instruction.isValid()) {
            instructions.add(instruction);
        }
    }

    /**
     * Adds a new instruction with the given opcode and operand to the given index of the sequence.
     *
     * @param index     Index at which the instruction has to be inserted.
     * @param opcode    Instruction's opcode.
     * @param operand   Instruction's operand.
     */
    public void insertInstruction(final int index, final Instruction.OpCode opcode,
                                  final short operand) {
        insertInstruction(index, new Instruction(opcode, operand));
    }

    // TODO: Document.
    public void insertRegisterInstruction(final int index,
                                          final RegisterInstruction.RegisterOpCode opcode,
                                          final short operand) {
        insertInstruction(index, new RegisterInstruction(opcode, operand));
    }

    /**
     * Adds the given instruction to the given index of the sequence.
     *
     * @param index         Index at which the instruction has to be inserted.
     * @param instruction    Instruction to insert into the sequence.
     */
    public void insertInstruction(final int index, final Instruction instruction) {
        if (instruction != null && instruction.isValid()) {
            instructions.add(index, instruction);
        }
    }

    /**
     * Removes the instruction at the given index from the sequence. If the index is outside the
     * sequence, no action is taken.
     *
     * @param index Index of the sequence to remove.
     */
    public void removeInstruction(int index) {
        if (index >= 0 && index < instructions.size()) {
            instructions.remove(index);
        }
    }

    /**
     * Saves the sequence as XML file inside the application's private data location. The suffix
     * ".kartseq" will be added to the filename that is actually saved.
     *
     * @param filename  Filename for the sequence.
     * @param context   Android context (needed to open output stream).
     * @throws IOException
     */
    public void save(final String filename, final Context context) throws IOException {
        // Open the file output stream.
        FileOutputStream output = context.openFileOutput(filename + ".kartseq",
                Context.MODE_PRIVATE);

        // Write XML content.
        output.write("<KartSequence>\n".getBytes());
        for (Instruction instruction: instructions) {
            output.write(("    " + instruction.toXmlTag() + "\n").getBytes());
        }
        output.write("</KartSequence>\n".getBytes());

        // Sync changes.
        output.close();
    }

    /**
     * Tries to load a kart sequence from the given XML file. Note that the file suffix ".kartseq"
     * is automatically added to the file name.
     *
     * @param filename                  File name of the sequence to load.
     * @param context                   Android application context (needed to open input stream).
     * @return                          Loaded kart sequence.
     * @throws IOException              Thrown if the file does not exists or is invalid.
     * @throws XmlPullParserException   Thrown if there was an error parsing the file's XML content.
     */
    public static KartSequence load(final String filename, final Context context)
            throws IOException, XmlPullParserException {
        // Create an empty kart sequence.
        KartSequence sequence = new KartSequence();

        // Try to open the file.
        FileInputStream input = context.openFileInput(filename + ".kartseq");

        // Parse the XML file.
        XmlPullParser parser = XmlPullParserFactory.newInstance().newPullParser();
        parser.setInput(input, null);
        int eventType = parser.getEventType();
        while (eventType != XmlPullParser.END_DOCUMENT) {
            switch (eventType) {
                case XmlPullParser.START_TAG:
                    if ("Instruction".equals(parser.getName())) {
                        Instruction.OpCode opcode = null;
                        RegisterInstruction.RegisterOpCode registerOpcode = null;
                        Short operand = null;
                        for (int i = 0; i < parser.getAttributeCount(); ++i) {
                            if ("opcode".equals(parser.getAttributeName(i))) {
                                try {
                                    opcode = Instruction.OpCode.valueOf(parser.getAttributeValue(i));
                                } catch (IllegalArgumentException e) {
                                    e.printStackTrace();
                                    throw new IOException("Invalid opcode: " + opcode);
                                }
                            } else if ("operand".equals(parser.getAttributeName(i))) {
                                operand = Short.parseShort(parser.getAttributeValue(i));
                            } else if ("registerOpcode".equals(parser.getAttributeName(i))) {
                                registerOpcode = RegisterInstruction.RegisterOpCode.valueOf(
                                        parser.getAttributeValue(i));
                            }
                        }
                        if (opcode != null && operand != null) {
                            if (opcode == Instruction.OpCode.MODIFY_REGISTER &&
                                    registerOpcode != null) {
                                sequence.addRegisterInstruction(registerOpcode, operand);
                            } else {
                                sequence.addInstruction(opcode, operand);
                            }
                        }
                    }
                    break;
            }
            eventType = parser.next();
        }
        return sequence;
    }

    /**
     * Returns a list of the available kart sequence files. Note that the file suffix is removed
     * from the file names.
     *
     * @param c  Android application context (needed to enumerate the files).
     * @return Sequence files names that can be loaded, or an empty array if no sequence have been found.
     */
    public static String[] availableSequences(final Context c) {
        // Get list of all files with the extension ".kartseq".
        String[] sequences = c.getFilesDir().list(new FilenameFilter() {
            @Override
            public boolean accept(File file, String filename) {
                return filename.endsWith(".kartseq");
            }
        });

        // Remove suffixes.
        for (int i = 0; i < sequences.length; ++i) {
            sequences[i] = sequences[i].replace(".kartseq", "");
        }

        return sequences;
    }

    public void saveToBundle(final Bundle bundle) {
        String[] sequenceStrings = new String[instructions.size()];
        for (int i = 0; i < instructions.size(); ++i) {
            sequenceStrings[i] = instructions.get(i).toString();
        }
        bundle.putStringArray("sequence", sequenceStrings);
    }

    public static final KartSequence loadFromBundle(final Bundle bundle) {
        KartSequence sequence = new KartSequence();
        if (bundle != null) {
            String[] sequenceStrings = bundle.getStringArray("sequence");
            if (sequenceStrings != null) {
                for (String sequenceString : sequenceStrings) {
                    Instruction instruction = Instruction.fromString(sequenceString);
                    if (instruction != null && instruction.isValid()) {
                        sequence.addInstruction(instruction);
                    }
                }
            }
        }
        return sequence;
    }

    // Converts the sequence to a raw Kart sequence VM instruction array and returns it.
    int[] getKartVMInstructions() {
        int[] instructions = new int[this.instructions.size()];
        for (int i = 0; i < instructions.length; ++i) {
            instructions[i] = this.instructions.get(i).getKartVMInstruction();
        }
        return instructions;
    }

    // Instructions list.
    private final List<Instruction> instructions = new LinkedList<Instruction>();
}
