package ch.hevs.kart;

import android.app.Activity;
import android.content.Context;
import android.content.SharedPreferences;

import java.util.EnumSet;
import java.util.HashSet;
import java.util.Set;

import ch.hevs.utils.Limits;

/**
 * Groups all kart settings together and allows convenient kart setup using returning reference
 * listening and setter chaining techniques.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
public class KartSetup {
	/**
	 * Returns the drive motor control block PWM period. The period is defined as the value the 
	 * main FPGA clock is divided by.
	 * 
	 * @return PWM period.
	 */
	public int drivePwmPeriod() {
		return drivePwmPeriod;
	}
	
	/**
	 * Sets the drive motor control block PWM period. The period is defined as the value the 
	 * main FPGA clock is divided by. Defaults to 64.
	 * 
	 * @param period	PWM period.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup drivePwmPeriod(final int period) {
		drivePwmPeriod = Limits.limit(period, 1, 65535);
		if (listener != null) listener.parameterChanged(Parameter.DrivePwmPeriod);
		return this;
	}

	/**
	 * Returns the maximal speed a kart is allowed to drive with. Values are from 0 to 15.
	 *
	 * @return Maximal drive speed.
	 */
	public int driveMaxSpeed() { return driveMaxSpeed; }

	/**
	 * Sets the maximal speed a kart is allowed to drive with. Valid values are in 0...15.
	 *
	 * @param speed		Maximal allowed speed.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup driveMaxSpeed(final int speed) {
		driveMaxSpeed = Limits.limit(speed, 0, 15);
		if (listener != null) listener.parameterChanged(Parameter.DriveMaxSpeed);
		return this;
	}

	/**
	 * Returns the maximal step count at the opposite site of the stepper end contact switch.
	 * 
	 * @return Maximal stepper motor step count.
	 */
	public int steeringMaxPosition() {
		return steeringMaxPosition;
	}

	// Returns the stepper motor center position.
	int steeringCenterPosition() {
		return steeringMaxPosition / 2;
	}
	
	/**
	 * Sets the maximal step count of the steering motor. Default value is defined by Kart VHDL 
	 * design.
	 * 
	 * @param position	The value of the step counter at the opposite site of the end switch.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringMaxPosition(final int position) {
		steeringMaxPosition = Limits.limit(position, 0, 65535);
		if (listener != null) listener.parameterChanged(Parameter.SteeringMaxPosition);
		return this;
	}

    /**
     * Sets the steering motor central position. Note that this method actually calls the
     * method steeringMaxPosition with 2 times the center position.
     *
     * @param position	The value of the step counter at the center.
     * @return			Reference to itself in order to chain the setters.
     */
    public KartSetup steeringCenterPosition(final int position) {
        return steeringMaxPosition(2 * position);
    }
	
	/**
	 * Returns the steering motor driver step period. The period is defined as the value the 
	 * main FPGA clock is divided by.
	 * 
	 * @return Steering motor step period.
	 */
	public int steeringStepPeriod() {
		return steeringStepPeriod;
	}
	
	/**
	 * Sets the steering motor driver step period. The period is defined as the value the 
	 * main FPGA clock is divided by. Defaults to 100.
	 * 
	 * @param period 	Steering motor step period.
	 * @return		 	Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringStepPeriod(final int period) {
		steeringStepPeriod = Limits.limit(period, 1, 65535);
		if (listener != null) listener.parameterChanged(Parameter.SteeringStepPeriod);
		return this;
	}
	
	/**
	 * Returns the steering end switch address from which emulated end switch the data is 
	 * considered by the steering FSM. This value can be ignored if the option to enable the
	 * emulated end contact is not set in the hardware settings.
	 * 
	 * @return Steering end contact address.
	 */
	public int steeringEndSwitchAddress() {
		return steeringEndSwitchAddress;
	}
	
	/**
	 * Sets the steering end switch address from which emulated end switch the data is 
	 * considered by the steering FSM. This value will be ignored if the option to enable the
	 * emulated end contact is not set in the hardware settings. Defaults to 0.
	 * 
	 * @param address	Steering end contact address.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringEndSwitchAddress(final int address) {
		steeringEndSwitchAddress = Limits.limit(address, 0, 31);
		if (listener != null) listener.parameterChanged(Parameter.SteeringEndSwitchAddress);
		return this;
	}
	
	/**
	 * Returns the Kart's actual hardware settings. Default is all flags cleared.
	 * 
	 * @return Kart's hardware settings.
	 */
	public EnumSet<KartHardwareSettings> hardwareSettings() {
		return hardwareSettings;
	}

	/**
	 * Changes all of the hardware settings flags to the given values.
	 *
	 * @param hardwareSettings	Hardware settings to apply.
	 */
	public KartSetup hardwareSettings(EnumSet<KartHardwareSettings> hardwareSettings) {
		this.hardwareSettings.clear();
		this.hardwareSettings.addAll(hardwareSettings);
		if (listener != null) listener.parameterChanged(Parameter.HardwareSettings);
		return this;
	}

	/**
	 * Sets the given hardware setting flag.
	 * 
	 * @param setting	Hardware setting flag to set.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup setHwSetting(final KartHardwareSettings setting) {
		hardwareSettings.add(setting);
		if (listener != null) listener.parameterChanged(Parameter.HardwareSettings);
		return this;
	}
	
	/**
	 * Clears the given hardware setting flag.
	 * 
	 * @param setting	Hardware setting flag to clear.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup clearHwSetting(final KartHardwareSettings setting) {
		hardwareSettings.remove(setting);
		if (listener != null) listener.parameterChanged(Parameter.HardwareSettings);
		return this;
	}
	
	/**
	 * Returns the update interval of the Kart in seconds. This interval is the time between each
	 * register update send from the Kart to the Android application.
	 * 
	 * @return Update interval in seconds.
	 */
	public float updateInterval() {
		return (float)updateInterval / 1000f;
	}
	
	// Returns the update interval in milliseconds.
	public int updateIntervalMs() {
		return updateInterval;
	}
	
	/**
	 * Sets the update interval of the Kart in seconds. This interval is the time between each
	 * register update send from the Kart to the Android application. Note that fractions of 
	 * seconds is supported too. Defaults to 0.064.
	 * 
	 * @param interval	Update interval in seconds.
	 * @return          Reference to itself in order to chain the setters.
	 */
	public KartSetup updateInterval(final float interval) {
		updateInterval = (int)(Limits.limit(interval, 0.001f, 65f) * 1000f);
		if (listener != null) listener.parameterChanged(Parameter.UpdateInterval);
		return this;
	}

    /**
     * Applies all parameters of the given Kart setup to this setup.
     *
     * @param setup Kart setup to apply to this kart setup object.
     * @return      Reference to itself in order to chain the commands.
     */
    public KartSetup apply(KartSetup setup) {
        drivePwmPeriod(setup.drivePwmPeriod);
		driveMaxSpeed(setup.driveMaxSpeed);
        steeringMaxPosition(setup.steeringMaxPosition);
        steeringStepPeriod(setup.steeringStepPeriod);
        steeringEndSwitchAddress(setup.steeringEndSwitchAddress);
        hardwareSettings(setup.hardwareSettings);
        updateInterval(setup.updateInterval());
        return this;
    }

    /**
     * Creates a new, kart setup with all parameters set to their defaults.
     */
    public KartSetup() {
        drivePwmPeriod = DefaultDrivePwmPeriod;
		driveMaxSpeed = DefaultDriveMaxSpeed;
        steeringMaxPosition = DefaultSteeringMaxPosition;
        steeringStepPeriod = DefaultSteeringStepPeriod;
        steeringEndSwitchAddress = DefaultSteeringEndSwitchAddress;
        hardwareSettings = EnumSet.noneOf(KartHardwareSettings.class);
        updateInterval = DefaultUpdateInterval;
        listener = null;
    }

	/**
	 * Saves the kart settings under the given name.
	 *
	 * @param setupName	Name for the saved setup.
	 * @param activity	Activity is needed to access file system.
     */
	public void save(String setupName, Activity activity) {
		SharedPreferences.Editor prefs =
				activity.getSharedPreferences(setupName, Context.MODE_PRIVATE).edit();
		prefs.putInt("DrivePwmPeriod", drivePwmPeriod);
		prefs.putInt("DriveMaxSpeed", driveMaxSpeed);
		prefs.putInt("SteeringMaxPosition", steeringMaxPosition);
		prefs.putInt("SteeringStepPeriod", steeringStepPeriod);
		prefs.putInt("SteeringEndSwitchAddress", steeringEndSwitchAddress);

		Set<String> hardwareSettingStrings = new HashSet<String>();
		for (KartHardwareSettings setting: hardwareSettings) {
			hardwareSettingStrings.add(setting.toString());
		}
		prefs.putStringSet("HardwareSettings", hardwareSettingStrings);

		prefs.putInt("UpdateInterval", updateInterval);

		prefs.apply();
	}

	/**
	 * Loads the kart settings from the given name.
	 *
	 * @param setupName	Name for the setup to load.
	 * @param activity	Activity is needed to access file system.
	 */
	public static KartSetup load(String setupName, Activity activity) {
		SharedPreferences prefs = activity.getSharedPreferences(setupName, Context.MODE_PRIVATE);
		if (prefs.contains("DrivePwmPeriod")) {
            KartSetup setup = new KartSetup();
			setup.drivePwmPeriod(prefs.getInt("DrivePwmPeriod", DefaultDrivePwmPeriod));
			setup.driveMaxSpeed(prefs.getInt("DriveMaxSpeed", DefaultDriveMaxSpeed));
            setup.steeringMaxPosition(prefs.getInt("SteeringMaxPosition", DefaultSteeringMaxPosition));
            setup.steeringStepPeriod(prefs.getInt("SteeringStepPeriod", DefaultSteeringStepPeriod));
            setup.steeringEndSwitchAddress(prefs.getInt("SteeringEndSwitchAddress",
                    DefaultSteeringEndSwitchAddress));

            Set<String> hardwareSettingStrings = prefs.getStringSet("HardwareSettings", null);
            if (hardwareSettingStrings != null) {
                for (String hardwareSettingString: hardwareSettingStrings) {
                    try {
                        KartHardwareSettings setting =
                                KartHardwareSettings.valueOf(hardwareSettingString);
                        setup.setHwSetting(setting);
                    } catch (NullPointerException e) {
                    } catch (IllegalArgumentException e) {}
                }
            }

            setup.updateInterval = prefs.getInt("UpdateInterval", DefaultUpdateInterval);

            return setup;
		}

        return null;
	}

	// Enum to identify which setup parameter has changed.
	enum Parameter {
		DrivePwmPeriod,
		DriveMaxSpeed,
		SteeringMaxPosition,
		SteeringStepPeriod,
		SteeringEndSwitchAddress,
		HardwareSettings,
		UpdateInterval
	}

	// Kart class implements this interface to get notified about changes in his setup.
	interface KartSetupListener {
		void parameterChanged(Parameter parameter);
	}

	// Creates a new kart setup object with the default values.
	KartSetup(final KartSetupListener listener) {
        drivePwmPeriod = DefaultDrivePwmPeriod;
		driveMaxSpeed = DefaultDriveMaxSpeed;
        steeringMaxPosition = DefaultSteeringMaxPosition;
        steeringStepPeriod = DefaultSteeringStepPeriod;
        steeringEndSwitchAddress = DefaultSteeringEndSwitchAddress;
        hardwareSettings = EnumSet.noneOf(KartHardwareSettings.class);
        updateInterval = DefaultUpdateInterval;
        this.listener = listener;
	}

	private static int DefaultDrivePwmPeriod = 64;
	private static int DefaultDriveMaxSpeed = 15;
	private static int DefaultSteeringMaxPosition = 600;
    private static int DefaultSteeringStepPeriod = 100;
    private static int DefaultSteeringEndSwitchAddress = 4;
    private static int DefaultUpdateInterval = 64;

	// The attributes containing the actual setup...
	private int drivePwmPeriod;
	private int driveMaxSpeed;
	private int steeringMaxPosition;
	private int steeringStepPeriod;
	private int steeringEndSwitchAddress;
	private final EnumSet<KartHardwareSettings> hardwareSettings;
	private int updateInterval;

	// Listener to update upon a change.
	private final KartSetupListener listener;
}
