package ch.hevs.kart.comm;

import java.util.LinkedList;
import java.util.List;

import android.content.Context;
import android.os.AsyncTask;

import ch.hevs.kart.comm.discovery.Discovery;
import ch.hevs.kart.comm.discovery.DiscoveryEntity;
import ch.hevs.kart.comm.discovery.DiscoveryListener;
import ch.hevs.kart.comm.protocol.Protocol;
import ch.hevs.kart.comm.protocol.ProtocolListener;
import ch.hevs.kart.comm.transport.Transport;
import ch.hevs.kart.comm.transport.bluetooth.RfCommDiscovery;
import ch.hevs.kart.comm.transport.dummy.DummyKart;
import ch.hevs.kart.comm.transport.tcp.TcpDiscovery;

/**
 * This class offers the ability to scan for new Kart entities using all discovery modules
 * of all transport implementations that actually support discovering.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
public final class CommunicationSetup {
	
	/*** Discovery ********************************************************************************/
	
	/**
	 * Start the discovery process for Kart entities using the given listener and Android
	 * 	      context.
	 * 
	 * Note that the discovery can be only active globally once. So if a discovery has been 
	 * started, another can only be started if the current one has been stopped.
	 * 
	 * @param listener	Listener that will get informed about all entities discovered.
	 * @param context	The android context from which the discovery is started.
	 */
	@SuppressWarnings("UnusedReturnValue")
    public static boolean startDiscovery(final DiscoveryListener listener, final Context context) {
		// Listener has to point actually to an instance!
		if (listener == null) return false;
		
		if (discoveryInProgress) {
			listener.errorIndication(null, "Can not start discovery until a running has stopped.");
			return false;
		}
		
		// Lazy initialization of all discover implementations.
		if (discoveries.isEmpty()) {
			discoveries.add(new RfCommDiscovery());
		}
		
		// Start scan on all discovery objects.
		for (Discovery discovery: discoveries) {
			if(!discovery.startDiscovery(listener, context)) {
				final String error = String.format("Could not start discovery %s !", 
						discovery.getClass().getSimpleName());
				listener.errorIndication(discovery, error);
				discoveryInProgress = false;
				return false;
			}
		}
		
		discoveryInProgress = true;
		return true;
	}
	
	/**
	 * Stops the actual discovery process (Does nothing at all if no process is actually
	 * 		  running).
	 */
	public static void stopDiscovery() {
		// If no actual discovery is in progress, do nothing.
		if (!discoveryInProgress) return;
		
		// Stop scan on all discovery objects.
		for (Discovery discovery: discoveries) {
			discovery.stopDiscovery();
		}
		
		discoveryInProgress = false;
	}

	public static void enableTcpTransport() {
		for (Discovery discovery: discoveries) {
			if (discovery instanceof TcpDiscovery) {
				return;
			}
		}
		discoveries.add(new TcpDiscovery());
	}

	public static void enableDummyKart() {
		if (!discoveries.contains(DummyKart.getDiscovery())) {
			discoveries.add(DummyKart.getDiscovery());
		}
	}

	// List of all discovery implementations.
	private static final List<Discovery> discoveries = new LinkedList<Discovery>();
	
	// Is a discovery is actually in progress.
	private static boolean discoveryInProgress = false;
	
	
	/*** Connection setup *************************************************************************/
	
	/**
	 * Tries to connect to the given Kart entity and sets the protocol listener to listen
	 * 		  to Kart protocol events. Note that this is started asynchronous and the status of the 
	 * 		  connection setup is returned using the CommunicationSetupListener interface.
	 * 
	 * @param entity	The entity to connect to.
	 * @param listener	The protocol listener getting the events from the kart communication.
	 * @param communicationSetupListener	The communication setup listener.
	 */
	public static void connect(final DiscoveryEntity entity, final ProtocolListener listener,
							   final CommunicationSetupListener communicationSetupListener) {
		// Setup listener must be given!
		if (communicationSetupListener == null) return;
		
		// If there is no protocol listener, fail.
		if (listener == null) {
			communicationSetupListener.communicationSetupFailed("Missing protocol listener!");
			return;
		}
		
		// If the entity does not define a transport implementation class, fail.
		if (entity.getTransportImplementation() == null) {
			communicationSetupListener.
                    communicationSetupFailed("Entity does not define a transport!");
			return;
		}
		
		// If no protocol is supported, fail.
		// TODO: Ask user which protocol to use if there are actually more than one...
		if (entity.getSupportedProtocols().isEmpty()) {
			communicationSetupListener.
                    communicationSetupFailed("Entity does not support any protocol!");
			return;
		}

		// Establish the communication inside a new asynchronous task.
		(new AsyncTask<Object, String, Object[]>() {
			// Thread local listener object.
			CommunicationSetupListener asyncCommunicationSetupListener;
			
			@Override
			protected Object[] doInBackground(Object... params) {
				// Get objects from parameter array.
				DiscoveryEntity entity = (DiscoveryEntity)params[0];
				ProtocolListener listener = (ProtocolListener)params[1];
				asyncCommunicationSetupListener = (CommunicationSetupListener)params[2];
				int protocolIndex = (Integer)params[3];
				
				// Instantiate the transport object.
				Transport transport;
				try {
					transport = entity.getTransportImplementation().newInstance();
				} catch (Exception e) {
					e.printStackTrace();
					publishProgress(e.getMessage());
					return null;
				}

				// Try to connect.
				try {
					transport.connect(entity);
				} catch (Exception e) {
					e.printStackTrace();
					publishProgress(e.getMessage());
					return null;
				}
				
				// Instantiate the protocol object.
				Protocol protocol;
				
				// Instantiate the protocol selected.
				try {
					protocol = entity.getSupportedProtocols().get(protocolIndex).newInstance();
				} catch (Exception e) {
					e.printStackTrace();
					publishProgress("Unable to create a protocol instance!");
					transport.disconnect();
					return null;
				}
				
				// Return the objects involved in the connection setup.
				return new Object[]{protocol, transport, listener};
			}

			@Override
			protected void onProgressUpdate(String... msg) {
				// Call the communication setup listener error message handler.
				asyncCommunicationSetupListener.communicationSetupFailed(msg[0]);
		    }

			@Override
			protected void onPostExecute(Object[] params) {
				if (params != null) {
					// Get objects from result array.
					Protocol protocol = (Protocol)params[0];
					Transport transport = (Transport)params[1];
					ProtocolListener listener = (ProtocolListener)params[2];

					// Finish the connection setup.
					transport.setListener(protocol);
					protocol.initialize(transport, listener);
					
					// Call the listeners setup finished callback.
					asyncCommunicationSetupListener.communicationSetupSucceeded(protocol);
				}
		    }
		}).execute(entity, listener, communicationSetupListener, 0);
	}
}
