package ch.hevs.kart.comm.transport.dummy;

import android.content.Context;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import ch.hevs.kart.KartSequence;
import ch.hevs.kart.comm.discovery.Discovery;
import ch.hevs.kart.comm.discovery.DiscoveryListener;
import ch.hevs.kart.comm.protocol.Protocol;
import ch.hevs.kart.comm.protocol.kartv1.KartV1Protocol;
import ch.hevs.kart.comm.transport.TransportListener;
import ch.hevs.kart.comm.transport.TransportState;
import ch.hevs.utils.Limits;
import ch.hevs.utils.Timer;

public class DummyKart {
    /*** Singleton instance ***********************************************************************/
    private static DummyKart instance;

    public static DummyKart getInstance() {
        if (instance == null) {
            instance = new DummyKart();
        }
        return instance;
    }

    /*** Kart simulation **************************************************************************/
    private final int[] controlRegisters = new int[16];
    private final int[] statusRegisters = new int[16];

    private final List<KartSequence.Instruction> sequence =
            new ArrayList<KartSequence.Instruction>();
    private boolean sequenceRunning = false;
    private int sequencePlayPosition = 0;
    private final Handler sequenceHandler = new Handler(Looper.getMainLooper());
    private final Runnable sequenceStep = new Runnable() {
        @Override
        public void run() {
            if (sequencePlayPosition < sequence.size()) {
                KartSequence.Instruction instruction = sequence.get(sequencePlayPosition);

                Log.d("Sequencer: ", instruction.getOpCode().toString() + "(" + String.valueOf(instruction.getOperand()) + ")");

                switch (instruction.getOpCode()) {
                    case NOP:
                        sequenceHandler.post(sequenceStep);
                        break;

                    case SET_SPEED:
                        controlRegisters[1] = Limits.limit(instruction.getOperand(), (short)-15, (short)15);
                        sequenceHandler.post(sequenceStep);
                        break;

                    case SET_ANGLE:
                        controlRegisters[3] = instruction.getOperand();
                        sequenceHandler.post(sequenceStep);
                        break;

                    case DRIVE_LEDS:
                        controlRegisters[6] = instruction.getOperand();
                        sequenceHandler.post(sequenceStep);
                        break;

                    case SET_LED:
                        controlRegisters[6] |= 1 << Limits.limit(instruction.getOperand(), (short)0, (short)3);
                        sequenceHandler.post(sequenceStep);
                        break;

                    case CLEAR_LED:
                        controlRegisters[6] &= 0xFE << Limits.limit(instruction.getOperand(), (short)0, (short)3);
                        sequenceHandler.post(sequenceStep);
                        break;

                    case RUN_DISTANCE:
                        // TODO: Better simulation...
                        sequenceHandler.postDelayed(sequenceStep, instruction.getOperand() * 100);
                        break;

                    case RUN_FOR:
                        sequenceHandler.postDelayed(sequenceStep, instruction.getOperand() * 10);
                        break;

                    case GOTO_IF_REG_NZ:
                        sequencePlayPosition = instruction.getOperand() - 1;
                        sequenceHandler.post(sequenceStep);
                        break;

                    case END:
                        sequenceRunning = false;
                        sequencePlayPosition = 0;
                        return;
                }

                ++sequencePlayPosition;

            } else {
                sequenceRunning = false;
                sequencePlayPosition = 0;
            }
        }
    };

    private static class Overflow {
        public static int overflow16(int value) {
            while (value > 65535) {
                value = value - 65535;
            }
            return value;
        }
    }

    private Timer updater = new Timer() {
        @Override
        public void onTimeout() {
            float time = (float)System.currentTimeMillis() / 1000.0f;

            // Simulate HALL sensors by incrementing the register in function to the speed.
            int absoluteSpeed = Math.abs(controlRegisters[1]);
            statusRegisters[0] = Overflow.overflow16(statusRegisters[0] + absoluteSpeed);
            sendStatusRegister(0);
            statusRegisters[1] = Overflow.overflow16(statusRegisters[0] + absoluteSpeed);
            sendStatusRegister(1);

            // Simulate stepper end contact status bit.
            statusRegisters[2] = statusRegisters[3] == 0 ? 0x01 : 0x00;

            // Simulate position reached status bit.
            boolean steeringActive = statusRegisters[3] != controlRegisters[3];
            statusRegisters[2] |= steeringActive ? 0x00 : 0x02;
            sendStatusRegister(2);

            // Simulate steering by limiting the moving speed to the target position.
            statusRegisters[3] = statusRegisters[3] +
                    Limits.limit(controlRegisters[3] - statusRegisters[3], -15, 15);
            sendStatusRegister(3);

            // Stepper end from I/O board not used.
            statusRegisters[4] = 0;
            sendStatusRegister(4);

            // Simulate battery voltage drop by using absolute speed and steering state and some
            // randomness.
            statusRegisters[5] = 26000 - 50 * absoluteSpeed;
            if (steeringActive) {
                statusRegisters[5] -= 200;
            }
            statusRegisters[5] -= (int)(Math.random() * 200.0);
            sendStatusRegister(5);

            // Distance is a sinus with about 0.1 Hz frequency and full range amplitude.
            statusRegisters[6] = (int)((32768f + Math.sin(2 * Math.PI * time) / 10f) * 32768f);
            sendStatusRegister(6);

            // Update sequence status register.
            statusRegisters[7] = sequenceRunning ? 0x01 : 0x00;
            statusRegisters[7] += sequencePlayPosition << 1;
            sendStatusRegister(7);
        }
    };

    private void setControlRegister(int address, int value) {
        if (!sequenceRunning && address >= 0 && address < controlRegisters.length) {
            controlRegisters[address] = value;
        }

        switch (address) {
            case 7:
                if ((value & 0x01) != 0) {
                    sequenceHandler.removeCallbacksAndMessages(null);
                    sequenceRunning = false;
                    sequencePlayPosition = 0;
                    sequence.clear();
                }
                if ((value & 0x02) != 0) {
                    if (!sequenceRunning) {
                        sequenceRunning = true;
                        sequenceHandler.post(sequenceStep);
                    }
                } else {
                    if (sequenceRunning) {
                        sequenceHandler.removeCallbacksAndMessages(null);
                        sequenceRunning = false;
                        sequencePlayPosition = 0;
                    }
                }
                break;

            case 8:
                if (!sequenceRunning) {
                    byte rawOpcode = (byte) ((value & 0xF000) >> 12);
                    KartSequence.Instruction.OpCode opCode = null;
                    for (KartSequence.Instruction.OpCode oc: KartSequence.Instruction.OpCode.values()) {
                        if (oc.toByte() == rawOpcode) {
                            opCode = oc;
                        }
                    }
                    if (opCode != null) {
                        sequence.add(new KartSequence.Instruction(opCode, (short) (value & 0x0FFF)));
                    }
                }
                break;
        }
    }

    private void sendStatusRegister(int address) {
        if (transport != null && transport.listener != null) {
            ByteBuffer frame = ByteBuffer.allocate(5);
            frame.put((byte) 0x55);
            frame.put((byte) address);
            frame.putShort((short) statusRegisters[address]);
            frame.put((byte) 0xAA);
            frame.flip();

            transport.listener.dataIndication(transport, frame);
        }
    }

    /*** Discovery support ************************************************************************/
    private static Discovery discovery;

    public static class DiscoveryEntity implements ch.hevs.kart.comm.discovery.DiscoveryEntity {
        @Override
        public String getFriendlyName() {
            return "Dummy Kart";
        }

        @Override
        public String getDescription() {
            return "Virtual local Kart simulation";
        }

        @Override
        public String getUuid() {
            return "DUMMY";
        }

        @Override
        public Class<? extends Transport> getTransportImplementation() {
            return Transport.class;
        }

        @Override
        public List<Class<? extends Protocol>> getSupportedProtocols() {
            List<Class<? extends Protocol>> protocols =
                    new ArrayList<Class<? extends Protocol>>();
            protocols.add(KartV1Protocol.class);
            return protocols;
        }

        @Override
        public boolean equals(ch.hevs.kart.comm.discovery.DiscoveryEntity other) {
            return other == this;
        }
    }

    public static Discovery getDiscovery() {
        if (discovery == null) {
            discovery = new Discovery() {
                @Override
                public boolean startDiscovery(DiscoveryListener listener, Context context) {
                    if (listener != null) {
                        listener.entityFoundIndication(this, new DiscoveryEntity());
                    }
                    return true;
                }

                @Override
                public void stopDiscovery() {
                }
            };
        }

        return discovery;
    }

    /*** Transport support ************************************************************************/
    Transport transport;

    public static class Transport implements ch.hevs.kart.comm.transport.Transport {
        private TransportListener listener;

        public Transport() {
            DummyKart.getInstance().transport = this;
        }

        @Override
        public void connect(String connectionString) throws Exception {
            if (listener != null) {
                listener.stateChangeIndication(this, TransportState.Connected);
            }
            DummyKart.getInstance().updater.schedulePeriodically(100);
        }

        @Override
        public void connect(ch.hevs.kart.comm.discovery.DiscoveryEntity entity) throws Exception {
            connect("");
        }

        @Override
        public boolean isConnected() {
            return true;
        }

        @Override
        public void disconnect() {
            DummyKart.getInstance().updater.stop();
        }

        @Override
        public void setListener(TransportListener listener) {
            this.listener = listener;
        }

        @Override
        public boolean write(ByteBuffer buffer) {
            buffer.flip();
            if (buffer.remaining() == 5 && buffer.get() == 0x55) {
                int address = (int)buffer.get() & 0xFF;
                int value = (int)buffer.getShort() & 0xFFFF;
                DummyKart.getInstance().setControlRegister(address, value);
            }

            return false;
        }
    }
}
