package ch.hevs.kart.ui;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.os.Bundle;
import android.view.KeyEvent;
import android.view.View;
import android.view.ViewGroup;
import android.view.inputmethod.EditorInfo;
import android.view.inputmethod.InputMethodManager;
import android.widget.CompoundButton;
import android.widget.EditText;
import android.widget.SeekBar;
import android.widget.Switch;
import android.widget.TextView;
import android.widget.Toast;

import ch.hevs.kart.Kart;
import ch.hevs.kart.KartHardwareSettings;
import ch.hevs.kart.KartListener;
import ch.hevs.kart.KartSetup;
import ch.hevs.kart.R;
import ch.hevs.kart.comm.discovery.DiscoveryEntity;

/**
 * Basic activity for all kart control activities.
 * 
 * This base class initializes and maintains the connection to the kart and disposes the protected
 * member attribute "kart" in order to enable all subclasses the access to the connected Kart
 * object.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
public abstract class AbstractKartControlActivity extends Activity implements KartListener {
	
	// Kart intent category to search for (RCA=Remote Control Activity).
	static final String INTENT_CATEGORY = "ch.hevs.kart.RCA";
	
	// Name of the entity passed from the discovery activity to this activity.
	static final String EXTRAS_ENTITY = "KART_COMMUNICATION_ENTITY";
	
	@Override
	public void onCreate(final Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		// Try to get the discovered entity.
		entity = (DiscoveryEntity) getIntent().getExtras().get(EXTRAS_ENTITY);
		if (entity == null) {
			Toast.makeText(this, "Invalid Kart entity!", Toast.LENGTH_LONG).show();
			finish();
		}

        // Try to load the settings for that kart from the storage.
        KartSetup persistentSetup = KartSetup.load(entity.getUuid() + ".kartsetup", this);
        if (persistentSetup != null) {
            kart.setup().apply(persistentSetup);
        }
	}
	
	@Override
	public void onResume() {
		super.onResume();
		
		// Try to connect to the Kart.
		kart.connect(entity);
		kart.addKartListener(this);
		kart.resetSteering();
	}
	
	@Override
	public void onPause() {
		// Disconnect from the Kart.
		kart.disconnect();
		kart.removeKartListener(this);
		super.onPause();
	}

    /**
     * Sets the enabled flag on the given view and all its subviews.
     *
     * @param view      View to set enabled flag (and all subviews).
     * @param enabled   True in order to enable the view, false in order to disable view.
     */
	protected static void setEnabledForAllSubviews(final View view, final boolean enabled) {
	    view.setEnabled(enabled);
	    view.setFocusable(enabled);

	    if(view instanceof ViewGroup) {
	        ViewGroup vg = (ViewGroup) view;
	        for (int i = 0; i < vg.getChildCount(); i++)
	            setEnabledForAllSubviews(vg.getChildAt(i), enabled);
	    }
	}

    /**
     * Shows a pop-up menu that enables the user to adjust kart setup and save the settings as
     * default settings for the given kart. Those settings are then reloaded automatically each time
     * when a connection to the kart is issued.
     *
     * Note: The settings are applied in the onCreate() method, it is still possible to override
     * them in your implementation of onCreate() after having called the super method.
     */
	protected void showKartSetupPopup() {
        // Create the popup dialog.
		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		View setupView = this.getLayoutInflater().inflate(R.layout.kart_setup_popup, null);
		builder.setView(setupView);

        // Initialize maximal speed setting UI and handle changes.
        final SeekBar driveMaxSpeed = (SeekBar)setupView.findViewById(R.id.driveMaxSpeed);
        driveMaxSpeed.setProgress(kart.setup().driveMaxSpeed());
        driveMaxSpeed.setOnSeekBarChangeListener(new SeekBar.OnSeekBarChangeListener() {
            @Override
            public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {
                kart.setup().driveMaxSpeed(progress);
            }

            @Override public void onStartTrackingTouch(SeekBar seekBar) {}
            @Override public void onStopTrackingTouch(SeekBar seekBar) {}
        });

        // Drive PWM period setting.
		final EditText drivePwmPeriod = (EditText)setupView.findViewById(R.id.drivePwmPeriod);
		drivePwmPeriod.setText(String.valueOf(kart.setup().drivePwmPeriod()));
        drivePwmPeriod.setOnEditorActionListener(new TextView.OnEditorActionListener() {
			@Override
			public boolean onEditorAction(TextView v, int actionId, KeyEvent event) {
				if (actionId == EditorInfo.IME_ACTION_DONE) {
					kart.setup().drivePwmPeriod(Integer.valueOf(drivePwmPeriod.getText().toString()));
                    ((InputMethodManager)getSystemService(Context.INPUT_METHOD_SERVICE)).hideSoftInputFromWindow(v.getWindowToken(), 0);
                    return true;
				}
                return false;
			}
		});

        // Drive inversion switch.
		final Switch inverseDriveMotorDirection = (Switch)setupView.findViewById(R.id.inverseDriveMotorDirection);
		inverseDriveMotorDirection.setChecked(kart.setup().hardwareSettings().contains(KartHardwareSettings.InverseDriveMotorDirection));
		inverseDriveMotorDirection.setOnCheckedChangeListener(new CompoundButton.OnCheckedChangeListener() {
			@Override
			public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
				if (inverseDriveMotorDirection.isChecked()) {
					kart.setup().setHwSetting(KartHardwareSettings.InverseDriveMotorDirection);
				} else {
					kart.setup().clearHwSetting(KartHardwareSettings.InverseDriveMotorDirection);
				}
			}
		});

        // Steering center position setting.
		final SeekBar steeringCenterPosition = (SeekBar)setupView.findViewById(R.id.steeringMaxPosition);
		steeringCenterPosition.setProgress(kart.setup().steeringMaxPosition() / 2);
		steeringCenterPosition.setOnSeekBarChangeListener(new SeekBar.OnSeekBarChangeListener() {
			@Override
			public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {
				kart.setup().steeringCenterPosition(progress);
			}

			@Override public void onStartTrackingTouch(SeekBar seekBar) {}
            @Override public void onStopTrackingTouch(SeekBar seekBar) {}
		});

        // Steering PWM period setting.
		final EditText steeringStepPeriod = (EditText)setupView.findViewById(R.id.steeringStepPeriod);
		steeringStepPeriod.setText(String.valueOf(kart.setup().steeringStepPeriod()));
        steeringStepPeriod.setOnEditorActionListener(new TextView.OnEditorActionListener() {
			@Override
			public boolean onEditorAction(TextView v, int actionId, KeyEvent event) {
				if (actionId == EditorInfo.IME_ACTION_DONE) {
					kart.setup().steeringStepPeriod(Integer.valueOf(steeringStepPeriod.getText().toString()));
                    ((InputMethodManager)getSystemService(Context.INPUT_METHOD_SERVICE)).hideSoftInputFromWindow(v.getWindowToken(), 0);
                    return true;
				}
				return false;
			}
		});

        // Inverse steering setting.
		final Switch inverseSteeringDirection = (Switch)setupView.findViewById(R.id.inverseSteeringDirection);
		inverseSteeringDirection.setChecked(kart.setup().hardwareSettings().contains(KartHardwareSettings.InverseSteeringDirection));
		inverseSteeringDirection.setOnCheckedChangeListener(new CompoundButton.OnCheckedChangeListener() {
			@Override
			public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
				if (inverseSteeringDirection.isChecked()) {
					kart.setup().setHwSetting(KartHardwareSettings.InverseSteeringDirection);
				} else {
					kart.setup().clearHwSetting(KartHardwareSettings.InverseSteeringDirection);
				}
			}
		});

        // Inverse steering end position switch side setting.
		final Switch inverseSteeringEndContactPosition = (Switch)setupView.findViewById(R.id.inverseSteeringEndContactPosition);
		inverseSteeringEndContactPosition.setChecked(kart.setup().hardwareSettings().contains(KartHardwareSettings.InverseSteeringEndContactPosition));
        inverseSteeringEndContactPosition.setOnCheckedChangeListener(new CompoundButton.OnCheckedChangeListener() {
            @Override
            public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
                if (inverseSteeringEndContactPosition.isChecked()) {
                    kart.setup().setHwSetting(KartHardwareSettings.InverseSteeringEndContactPosition);
                } else {
                    kart.setup().clearHwSetting(KartHardwareSettings.InverseSteeringEndContactPosition);
                }
            }
        });

        // Reset steering button.
		setupView.findViewById(R.id.resetSteering).setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				kart.resetSteering();
			}
		});

        // Save settings persistent button.
        setupView.findViewById(R.id.saveDefault).setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                kart.setup().save(entity.getUuid() + ".kartsetup", AbstractKartControlActivity.this);
            }
        });

        // Show the dialog.
		builder.create().show();
	}

	// The entity describing the connection to the Kart.
	private DiscoveryEntity entity;

    /**
     * The Kart object. You should use this object for all communication with the kart.
     */
	protected final Kart kart = new Kart();
}
